//-----------------------------------------------------------------------------
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Grace.cc: --- communicate with runnig Grace process via pipes
//
//	CHANGE HISTORY:
//		MB, 1999-??-??: 1.0.0: - first version; uses xmgr
//		UL, 1999-10-25: 1.1.0: - changed to grace
//		UL, 2000-02-08: 1.2.0: - add support for all 1-D region types
//								 (fixes SPR 1086 on xmmselct: task crashes
//								  when trying to read in 1D region
//		UL, 2000-07-10: 1.3.0: - reimplemented Grace::isUp() using
//								 GraceIsOpen(); this should resolve a problem
//								 with SIGPIPE exceptions in xmmselect
//								 (after uses has closed Grace window with
//								  "Exit" button)
//		UL, 2000-12-22: 1.3.1: - trivial change to make optimizer happy
//
//	AUTHORS:
//		Marco Beijersbergen (MB)
//		Uwe Lammers (UL)
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/dsplot/lib/grace/Grace.cc,v 1.18 2009/08/31 16:14:47 aibarra Exp $
//
#include <cstdlib>
#include <unistd.h>

#include <Grace.h>
#include <grace_np.h>
#include <fstream>
#include <sstream>
#include <map>
#include <algorithm>

#include "errstr.h"
#include "common.h"
#include "stringutil.h"

using namespace std;

static bool graceError = false;
static void graceErrorFunc(const char *)
{
	graceError = true;
}

Grace::Grace(const string &s)
{
	//	open pipe of size 1024 to Grace subprocess in nosafe-mode to allow
	//	modification to local files in batch mode
	/* AI gcc4 porting
	* error: deprecated conversion from string constant to ‘char*’
	* => casting (char *)
	*/
	GraceOpenVA((char*) "xmgrace", 1024, (char*) "-nosafe", (char *) "-noask", NULL);

	if (s.length())
		readSets(s);

	GraceRegisterErrorFunction(graceErrorFunc);
}


Grace::~Grace()
{
	GraceFlush();
	GraceClose();
}


bool Grace::isUp() const
{
	
//	GracePrintf("");
//
//	bool res = !graceError;
//	graceError = false;

//	return res;
	return GraceIsOpen();
} // Grace::isUp()


void Grace::readSets(const string &name)
{
	GracePrintf("read \"%s\"", name.c_str());
	GracePrintf("autoscale");
}

void Grace::writeSet(int setNr, const string &fileName) const
{
	GracePrintf("write s%d file \"%s\"", setNr,fileName.c_str());
}

void Grace::killAll()
{
	GracePrintf("kill s0");
}

void Grace::killSet(int setNr)
{
	GracePrintf("kill %d", setNr);
}


void Grace::getRegions(const string &parameterFile,
					   vector<string>& regions) const
{
  typedef map<int,string,less<int> > RegionMap;
  RegionMap regionMap;

  ifstream ifs(parameterFile.c_str());
//
//	UL: The followig check for existence of ifs causes the method to
//		always return without having read anything from the parameter
//		file - reason unknown - some synchronisation problem in the
//		grace library - to be further investigated
//  if (!ifs) {
//    errstr << "cannot read parameter file " << parameterFile << Sas::warning("readParameterFile");
//    return;
//  }

	string line;
	while (getline(ifs,line)) {
		string l = trim(line, true, "@ \t");
		if (l[0] == 'r') {
			int regionNr;
			stringstream s;
			s<<l.substr(1,string::npos);
			s>>regionNr;
			if (!s) continue;
			regionMap[regionNr] += l.substr(3,string::npos) + ';';
		}
	}

	for (RegionMap::const_iterator i=regionMap.begin();
													i!=regionMap.end(); ++i)
		regions.push_back((*i).second);
}


vector<string> Grace::regionSelection() const
{
	/* AI gcc4 porting
	* error: deprecated conversion from string constant to ‘char*’
	* => casting (char *)
	*/
	char *parameterFile = (char *) "tmp.param";
	GracePrintf("saveall \"%s\"", parameterFile);
	vector<string> regions;
	getRegions(parameterFile,regions);
	unlink(parameterFile);
	return regions;
}


Grace::RegionType Grace::type(const vector<string> &comp)
{
	/* AI gcc4 porting
	* error: deprecated conversion from string constant to ‘char*’
	* => casting (char *)
	*/
	static char *_s[] = { (char *)"above", (char *)"below",(char *) "left", (char *)"right",
						 (char *) "horizi", (char *)"horizo", (char *)"verti",(char *) "verto",
						 (char *) "polyi",(char *) "polyo" };
	static vector<string> s(_s, _s + NUM_OF(_s));

	RegionType reg = NONE;
	bool on = false;
	for (unsigned n=0; n<comp.size(); ++n) {
		vector<string> v(breakup(comp[n], " ,"));
		if (v.size() < 1)
			continue;
		if (v[0] == "type") {
			vector<string>::const_iterator i = ::find(s.begin(), s.end(),
													  v[1]);
			if (i != s.end())
				reg = static_cast<RegionType>(i-s.begin());
			else
				throw GeoRegion::InvalidRegion();
		} else if (uppercase(v[0]) == "ON")
			on = true;
	}

	return on && reg != NONE ? reg : NONE;
} // DsplotRegionSpecification::type()


GeoRegion Grace::toRegion(const vector<string> &regionSpec)
{
	PROC

	GeoRegion		region;
	region.push_back(Component());
	Component	&comp = region[0];

	for (unsigned n=0; n<regionSpec.size(); ++n) {
		string s(regionSpec[n]);	//	make a copy

		//	chop if everything after '#'
		string::size_type pos;
		if ((pos = s.find("#")) != string::npos)
			s=trim(s.substr(0, pos));

		//	skip empty lines
		if (s.length() == 0)
			continue;

		vector<string> fields(breakup(s, ";"));
		if (fields.size() < 3)	//	we need at least that many fields for
								//	a region
			continue;

		#define THROWINVALID(minsize)	\
			if (fields.size() < minsize) { throw Region::InvalidRegion(); }

		RegionType whatType = type(fields);
		if (whatType != INSIDE_POLY && whatType != OUTSIDE_POLY &&
			whatType != NONE) {
			double x1 = 0., y1 = 0., x2 = 0., y2 = 0.;
			unsigned n;
			for (n=0; n<fields.size(); ++n) {
				vector<string> v(breakup(fields[n], " ,"));
				if (v.size() == 5 && v[0] == "line") {
					//	found it
					x1 = toNumber<double>(v[1]);
					y1 = toNumber<double>(v[2]);
					x2 = toNumber<double>(v[3]);
					y2 = toNumber<double>(v[4]);
					break;
				}
			}
			if (n >= fields.size())
				throw GeoRegion::InvalidRegion();

			FShape *line;
			switch(whatType) {
				case ABOVE_LINE:
				case BELOW_LINE:
				case IN_HORIZ_RANGE:
				case OUT_HORIZ_RANGE:
				case IN_VERTI_RANGE:
				case OUT_VERTI_RANGE:
					line = new FLine(FPoint(x1, y1), FPoint(x2, y1));
					break;
				case LEFT_OF_LINE:
					line = new FLine(FPoint(x1, min(y1, y2)),
									 FPoint(x2, max(y1, y2)));
										//	y-coordinate of first point
										//	lower than second point signal
										//	LEFT_OF_LINE
					break;
				default: // RIGHT_OF_LINE
					line = new FLine(FPoint(x1, max(y1, y2)),
									 FPoint(x2, min(y1, y2)));
										//	y-coordinate of second point
										//	lower than second point signal
										//	RIGHT_OF_LINE
					break;
			}
			line->include(whatType != OUT_HORIZ_RANGE &&
						  whatType != OUT_VERTI_RANGE);
			comp.push_back(line);			
		} else if (whatType == INSIDE_POLY || whatType == OUTSIDE_POLY) {
			//
			//	we have a polygon
			//
			vector<float> x, y;
			for (unsigned n=1; n<fields.size(); ++n) {
				vector<string> s(breakup(fields[n], " ,"));
				if (s.size() != 3)
					continue;
				if (s[0] != "xy")
					continue;
				x.push_back(toNumber<float>(s[1]));
				y.push_back(toNumber<float>(s[2]));
			}

			if (x.size() < 3)
				//	cannot have a polygon with less than 3 points
				throw GeoRegion::InvalidRegion();

			FShape *poly = new FPolygon(x, y, x.size());
			poly->include(whatType == INSIDE_POLY);
			comp.push_back(poly);
		}
	}

	RETURN region;
} // Grace::toRegion()
