//-----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	  dstoplot.cc	--- convert Dal data set to Grace format
//
//	DESCRIPTION:
//	  A task subroutines that converts the contents of a table in a
//	  data set into a format that can be ingested by the external plotting
//	  program Grace.
//
//	CHANGE HISTORY:
//		MB, 1999-??-??: 1.0.0: - first version
//		UL, 1999-10-11: 1.1.0: - added ConvertableData::mode() to bring
//								 in line with dal-1.74
//							   - fixed SPR 760 : insufficient precsision in
//								 dstoplot output
//		UL, 1999-10-24: 1.2.0: - fixed _really_ SPR 760; real data
//								 are written with 8/16 significant digits
//								 depending on their type (real32/real64)
//							   - added support for scatter plots;
//								 parameter "withscatter"
//							   - bail out with error on vector columns (this
//								 fixes SPR 641)
//							   - implemented 'withoffsetx' parameter:
//									x[i] -= x[0]
//								 for plots of rate curves with large time tags
//		UL, 2000-01-07: 1.3.0: - fixed SPR 1030 : xmgrace incompatible
//								 (way legends are added in grace is different
//								  from xmgr)
//		UL, 2000-02-24: 1.4.0: - added dummy ConvertibleData::release()
//								 since dal has changed in v1.104
//		UL, 2000-05-09: 1.5.0: - provide x-axis offset as output parameter
//		UL, 2000-07-12: 2.0.0: - general overhaul in the course of implementing
//								 support for vector columns
//		UL, 2000-07-21: 2.1.0: - fixed segfault problem caused by deleting
//								 ColumnData * in destructor *after* data set
//								 had been closed
//							   - reimplement pre-1.10-behavior: if the table
//								 contains only one numerical column this
//								 is plotted as Y against index as X
//		UL, 2000-11-16: 2.2.0: - check for NaNs in input table
//								 (this fixes SPR 1853: Cannot plot RGS net
//								  spectral files)
//		UL, 2000-12-20: 2.3.0: - new parameters:
//									xmin/xmax/ymin/ymax/xlog/ylog
//		UL, 2002-01-24: 2.4.0: - added support for error bars
//		UL, 2002-02-14: 2.5.0: - better robustness against invalid row-range
//								 specifiers
//		UL, 2003-01-17: 2.6.0: - added support for plotting abscissa ranges
//								 (e.g. GTIs in time-series plots) separatly
//								 or overlaid onto normal plot
//		UL, 2003-01-21: 2.7.0: - several stdc++ compliance fixes enforced by
//								 gcc-3.2
//		UL, 2003-03-14: 2.8.0: - added parameters
//									o withalltablesafter
//									o subtitle
//
//	AUTHORS:
//		Marco Beijersbergen (MB)
//		Uwe Lammers (UL)
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/dsplot/src/dstoplot.cc,v 1.36 2009/08/31 16:14:47 aibarra Exp $
//
#include <algorithm>
#include <iomanip>	//	setprecision
#include <fstream>
#include <cstdlib>
#include<cassert>

#include "param.h"
#include "Dal.h"
#include "SetTableColumn.h"
#include "Filename.h"
#include "stringutil.h"
#include "NAN.h"
#include "errstr.h"

#include "dstoplot.h"

/*
 * Possible work around for Panther.
#ifdef MACOSX
 static long isnan(double value) { return __isnand(value); }
#endif
*/

using namespace std;
//
//	--------------------------( NanOp node )---------------------------------
//
extern "C" int isnan(double);
	// work around bug in cmath of gcc-3.1 on MacOSX where isnan is a macro
	// that is #undef'ed, hence the follwing line leads to a
	//	missing-isnan-declaration problem; on all other platform the extra
	//	external-declaration is harmless

//
//	A class that represents a data vector of a fixed length; it has
//	a name, size,  and its values can be inquired as double precision numbers
//
class FixedSizeDataVector {
public:
	FixedSizeDataVector(const string &name, unsigned size = 0)
		: _name(name), _size(size) {}

	const string &name() const { return _name; }
	unsigned size() const { return _size; }

	virtual string units() const { return ""; }
	virtual bool isDouble() const { return false; }
	virtual double value(unsigned row) const = 0;
	virtual double min() const = 0;
	virtual double max() const = 0;

	virtual ~FixedSizeDataVector() {}

protected:
	friend class PlotableDataVector;
	void size(unsigned n) { _size = n; }

private:
	string _name;
	unsigned _size;
};

//
//	A class that represents a FixedSizeDataVector represented as
//	data columns in a Dal table.
//
class DalFixedSizeDataVector : public FixedSizeDataVector {
public:
	//
	//	construct from pointer to table and a name which can point to
	//	a scalar or a vector column; in the latter case it can be
	//	suffixed with "[n]" where <n> is a number >=0 and less than
	//	the length of the vector column
	//
	DalFixedSizeDataVector(const Table *table,
						   const string &columnName)
		: FixedSizeDataVector(columnName, table->rows()), _minMaxSet(false)
	{
		vector<string> v(breakup(columnName, "[]"));
		string colName = v[0];
		_cd = table->column(colName)->data();
		_scalar = _cd->elements() == 1;
		_type = _cd->dataType();
		switch (_type) {
			case Dal::Bool	: _data = _cd->boolData(); break;
			case Dal::Int8	: _data = _cd->int8Data(); break;
			case Dal::Int16	: _data = _cd->int16Data(); break;
			case Dal::Uint16: _data = _cd->uint16Data(); break;
			case Dal::Int32	: _data = _cd->int32Data(); break;
			case Dal::Uint32: _data = _cd->uint32Data(); break;
			case Dal::Real32: _data = _cd->real32Data(); break;
			case Dal::Real64: _data = _cd->real64Data(); break;
			default:
				errstr<<"Column `"<<colName<<"' is non-numeric"
					  <<Sas::error("NonNumericColumn");
		}

		if (_scalar) {
			if (v.size() != 1)
				errstr<<"Column `"<<colName<<"' is scalar - ignoring row "
						"specifier"<<Sas::warning("IllegalRowSpecifier");
		} else {
			//	vector column
			size(_cd->elements());
			_row = v.size()>=2 ? toNumber<unsigned>(v[1]) : 0;
			if (_row>=table->rows())
				errstr<<"Table has only "<<table->rows()<<" rows - illegal "
					  <<"row specifier "<<_row
					  <<Sas::error("IndexOutOfBounds");
			else if (v.size()==0 || v.size()>2)
				errstr<<"`"<<columnName<<"' is not a valid data vector "
						"specifier"
					  <<Sas::error("InvalidDataColumnSpecification");
		}

		//	check dimension
		if (_cd->parent()->dimensions() > 2)
			errstr<<"Only scalar and vector columns supported `"
				  <<_cd->parent()->qualifiedName()<<"' is neither of this."
				  <<Sas::error("MultidimensionalColumn");
	}

	bool isDouble() const { return _type == Dal::Real64; }

	string units() const { return _cd->parent()->units(); }

	double value(unsigned row) const
	{
		double r = 0;

		if (row < size()) {
			unsigned ind = _scalar ? row : _cd->position(_row, row);
			switch (_type) {
				case Dal::Bool: r =
					reinterpret_cast<const bool8 *>(_data)[ind] ? 1. : 0.;
					break;
				case Dal::Int8:
					r = reinterpret_cast<const int8 *>(_data)[ind];
					break;
				case Dal::Int16:
					r = reinterpret_cast<const int16 *>(_data)[ind];
					break;
				case Dal::Uint16:
					r = reinterpret_cast<const uint16 *>(_data)[ind];
					break;
				case Dal::Int32:
					r = reinterpret_cast<const int32 *>(_data)[ind];
					break;
				case Dal::Uint32:
					r = reinterpret_cast<const uint32 *>(_data)[ind];
					break;
				case Dal::Real32: {
					const real32 *p = reinterpret_cast<const real32 *>(_data);
					if (NaN.isSet(p[ind]))
						NaN.set(r);
					else
						r = p[ind];
					}
					break;
				case Dal::Real64: {
					const real64 *p = reinterpret_cast<const real64 *>(_data);
					if (NaN.isSet(p[ind]))
						NaN.set(r);
					else
						r = p[ind];
					}
					break;
				default:
					break;
			}
		} else
			errstr<<"`"<<name()<<"' has only "<<size()<<" elements but "
					"position "<<row<<" requested - returning with 0"
				  <<Sas::warning("IndexOutOfBounds");

		return r;
	}

	double min() const {
		minmax();

		return _min;
	}

	double max() const {
		minmax();

		return _max;
	}

	~DalFixedSizeDataVector() { delete _cd; }

private:
	bool				_scalar;
	unsigned			_row;
	Dal::DataType		_type;
	const void			*_data;
	const ColumnData	*_cd;

	mutable bool		_minMaxSet;
	mutable double		_min, _max;

	void minmax() const {
		if (_minMaxSet)
			return;

		_min = 1.e300;	//	std::numeric_limits<double>::max()
		_max = -_min;

		unsigned N = size();
		for (unsigned n=0; n<N; ++n) {
			double val = value(n);
			_min = std::min(_min, val);
			_max = std::max(_max, val);
		}
	}
};

//
//	A class that is a simple index vector; having values from [0, size-1]
//
class IndexVector : public FixedSizeDataVector {
public:
	IndexVector(unsigned size) : FixedSizeDataVector("INDEX", size) {}
	double value(unsigned row) const { return static_cast<double>(row); }
	double min() const { return 0.; }
	double max() const { return size()-1.; }
	~IndexVector() {}
};

//
//	PlottableDataVector: vector with pointers to FixedSizeDataVectors;
//	first elements is xAxis, all subsequent element are y-Axis data
//	columns
//
namespace {

bool minCmp(const FixedSizeDataVector *a, const FixedSizeDataVector *b)
{
	return a->min() < b->min();
}

bool maxCmp(const FixedSizeDataVector *a, const FixedSizeDataVector *b)
{
	return a->max() < b->max();
}

}

class PlotableDataVector {
public:
	PlotableDataVector(const Table *table,
					   const string &xAxis, bool withIndex,
					   const vector<string> &yAxes)
	{
		//	collect all numeric columns
		vector<string> numColNames;
		for (unsigned n=0; n<table->columns(); n++)
			if (table->column(n)->dataType() != Dal::DString)
				numColNames.push_back(table->column(n)->name());

		//
		//	pick an x-axis
		//
		if (!withIndex)
			if (!xAxis.empty())
				addVector(table, xAxis, true);
			else
				if (!numColNames.empty())
					addVector(table, numColNames[0]);
				else {
					_data[0].push_back(new IndexVector(table->rows()));
					_data[1].push_back(0);
					_data[2].push_back(0);
				}
		else {
			_data[0].push_back(new IndexVector(table->rows()));
			_data[1].push_back(0);
			_data[2].push_back(0);
		}

		//
		//	select columns to plot as y
		//
		if (yAxes.size())
			for (unsigned n=0; n<yAxes.size(); addVector(table, yAxes[n++])) ;
		else {
			if (numColNames.empty())
				errstr<<"The table `"<<table->qualifiedName()<<"' contains "
						"no numeric columns"
					  <<Sas::error("NoNumericColumns");
			for (unsigned n=0; n<numColNames.size(); ++n)
				if (_data[0][0]->name() != numColNames[n])
					addVector(table, numColNames[n]);
		}

		if (size()==1) {
			//	there is only one numerical column (which is in use as
			//	x-axis); make this the y-axis and use index axis as x
			_data[0].insert(_data[0].begin(), new IndexVector(table->rows()));
			_data[1].insert(_data[1].begin(), 0);
			_data[2].insert(_data[2].begin(), 0);
		}

		adjustXsize();
	} // PlotableDataVector()


	//
	//	adjust size of X-axis to smallest value of y data axis extents
	//
	void adjustXsize() {
		unsigned N = ~0;
		for (unsigned n=1; n<size(); ++n)
			N = ::min(N, _data[0][n]->size());
		_data[0][0]->size(N);
	} // adjustXSize()


	//	return number of plotable data vectors
	unsigned size() const { return _data[0].size(); }

	//	check if vector has associated X/Y error data
	bool hasXerrors(unsigned i) const { return _data[1][validatedIndex(i)]; }
	bool hasYerrors(unsigned i) const { return _data[2][validatedIndex(i)]; }

	//	give access to data + error vectors
	const FixedSizeDataVector *data(unsigned i) const {
		return _data[0][validatedIndex(i)];
	}
	const FixedSizeDataVector *xError(unsigned i) const {
		return _data[1][validatedIndex(i)];
	}
	const FixedSizeDataVector *yError(unsigned i) const {
		return _data[2][validatedIndex(i)];
	}

	double min() const {
		return (*min_element(_data[0].begin() + 1,	//	skip X
				_data[0].end(), minCmp))->min();
	}

	double max() const {
		return (*max_element(_data[0].begin() + 1,	//	skip X
							 _data[0].end(), maxCmp))->max();
	}

	~PlotableDataVector() {
		unsigned N = size();
		for (unsigned n=0; n<N; ++n) {
			delete _data[0][n];
			delete _data[1][n];
			delete _data[2][n];
		}
	} // ~PlotableDataVector()

private:
	unsigned validatedIndex(unsigned i) const {
		if (i>=_data[0].size())
			errstr<<"Index value ("<<i<<") out of valid range "
					"[0, "<<_data[0].size()-1<<']'
				  <<Sas::error("IndexOutOfRange");

		return i;
	} // validatedIndex()

	void addVector(const Table *table, const string &compoundColName,
				   bool singleColumn = false)
	{
		//	dissect "colum.xerror.yerror"
		vector<string> v(breakup(compoundColName, "."));
		string colNames[3];
		if (v.size() >= 1 && v.size() <= 3 && v[0].length()>0) {
			colNames[0] = v[0];		//	data column name
			if (v.size() == 2)		//	column.yerror
				colNames[2] = v[1];
			if (v.size() == 3)
				colNames[1] = v[1], colNames[2] = v[2];
		} else
			errstr<<"Given data vector specification `"<<compoundColName
				  <<"' does not match expected pattern "
					"datacol|datacol.yError|datacol.xError.yError"
				  <<Sas::error("InvalidDataColumnSpecification");

		v = breakup(colNames[0], "[]");
		string colName = v[0];
		if (v.size() == 1) {
			_data[0].push_back(new DalFixedSizeDataVector(table, colName));
			for (unsigned n=1; n<=2; ++n)
				_data[n].push_back(!colNames[n].empty()
						? new DalFixedSizeDataVector(table, colNames[n]) : 0);
		} else {
			//	vector column
			v = breakup(v[1], ",:");
			assert(v.size()>0);
			unsigned from = 0, to = 0, steps = 1;
			try {
				from = to = std::max(toNumber<unsigned>(v[0]), 0U);
				if (v.size()>=2)
					to = std::max(toNumber<unsigned>(v[1]), from);
				if (v.size()>=3)
					steps = std::max(toNumber<unsigned>(v[2]), 1U);
			} catch (...) {
				errstr<<"`"<<v[1]<<"' is not a valid row-range specifier"
					  <<Sas::error("InvalidRowRangeSpecifier");
			}

			string errorNames;
			for (unsigned n=1; n<=2; ++n)
				if (!colNames[n].empty())
					errorNames += "." + colNames[n];
			for (unsigned n=from; n<=to; n+=steps) {
				_data[0].push_back(new DalFixedSizeDataVector(table,
								   colName + '[' + toString(n) + ']'));
				for (unsigned k=1; k<=2; ++k)
					_data[k].push_back(!colNames[k].empty()
						? new DalFixedSizeDataVector(table, colNames[k]) : 0);
				if (singleColumn)
					break;
			}
		}
	} // addVector()

	//	data members
	vector<FixedSizeDataVector *>	_data[3];	// 0: data values
												// 1: errors in X
												// 2: errors in Y
};

class AbscissaRanges {
public:
	AbscissaRanges(const vector<string> &tableList,
				   const string &fromCol, const string &toCol) :
		_tableList(tableList), _fromCol(fromCol), _toCol(toCol) {}

	void plot(double yBaseLine, double yStride,
			  bool withOffsetX, double offsetX,
			  unsigned plotVectorIndex, bool onlyRanges,
			  const string &plotTitle,
			  const string &plotSubTitle,
			  ostream &os) const {

		if (!plotTitle.empty())
			os<<"@ title "<<'"'<<plotTitle<<'"'<<endl;

		if (!plotSubTitle.empty())
			os<<"@ subtitle "<<'"'<<plotSubTitle<<'"'<<endl;

		for (unsigned n=0; n<_tableList.size(); ++n) {
			SetTableColumn stc(_tableList[n]);
			DataSet *set = stc.dataSet();
			//	keep first set in memory
			if (!n)
				dataSetServer->keep(set->name());
			
			Table *tab = stc.table(); {

			DalFixedSizeDataVector rangeFrom(tab, _fromCol),
								   rangeTo(tab, _toCol);

			if (withOffsetX) {
				if (offsetX == 0. && rangeFrom.size() > 0)
					offsetX = rangeFrom.value(0);
			} else
				offsetX = 0.;

			if (onlyRanges && !n) {
				os<<"@ xaxis label "<<'"'<<rangeFrom.name()<<'/'
				  <<rangeTo.name();
				if (withOffsetX)
					os<<" - "<<setprecision(8)<<offsetX;

				if (!rangeFrom.units().empty())
					os<<" ["<<rangeFrom.units()<<"]";
				os<<'"'<<endl;
			}

			unsigned &j = plotVectorIndex;
			os<<"@  s"<<j<<" line type 0\n";
			os<<"@	type xydx\n";
			unsigned N = rangeFrom.size();
			if (!n && N && onlyRanges)
				os<<rangeFrom.value(0)-offsetX<<' '<<yBaseLine-.5*yStride
				  <<" 0\n";
			for (unsigned k=0; k<N; ++k) {
				double x0 = rangeFrom.value(k),
					   x1 = rangeTo.value(k);
				os<<.5*(x0+x1)-offsetX<<' '<<yBaseLine+n*yStride<<' '
				  <<.5*(x1-x0)<<endl;
			}
			if (n==_tableList.size()-1 && N && onlyRanges)
				os<<rangeTo.value(N-1)-offsetX<<' '<<yBaseLine+10.*yStride
				  <<" 0\n";
			os<<"&"<<endl;
			++j;
			}
			dataSetServer->close(set);
		}
	}

private:
	vector<string>	_tableList;
	string			_fromCol, _toCol;
};

//
//	The task subroutine.
//
void dstoplot()
{
	// setup output channel
	ostream* os = &cout;

	ofstream ofs;
	if (stringParameter("output") == "file") {
		string outputFile = stringParameter("outputfile");
		ofs.open(outputFile.c_str());
		if (!ofs)
			errstr<<"could not open "<<outputFile<<Sas::error("openFile");
			os = &ofs;
	}

	*os<<"# Grace project file\n";
	*os<<"#\n";
	
	//AI (SOC-SPR-2536)
	//New versions of xmgrace need the keyword on the header to
	//plot the graphs properly.
	*os<<"@ version 40102\n";
	double xWorldMin = 0.,
		   xWorldMax = 0.,
		   yWorldMin = 0.,
		   yWorldMax = 0.;
	bool offsetX = booleanParameter("withoffsetx"),
		 withTable = booleanParameter("withtable");
	double xOffset = 0;
	unsigned plotIndex = 0;
	string plotTitle = stringParameter("title"),
		   plotSubTitle = stringParameter("subtitle");

	if (withTable) {
		SetTableColumn stc(stringParameter("table"));
		DataSet *set = stc.dataSet(); {
		const Table *table = stc.table();
		PlotableDataVector plotVec(table,
								   stringParameter("x"),
								   booleanParameter("index"),
								   stringListParameter("y"));
		yWorldMin = plotVec.min();
		yWorldMax = plotVec.max();

		const FixedSizeDataVector *xCol = plotVec.data(0);
		xWorldMin = xCol->min();
		xWorldMax = xCol->max();

		//	offset x-axis values
		if (offsetX && table->rows()) {
			xOffset = xCol->value(0);
			setParameter("offsetx", xOffset);
		}

		unsigned xprec = 8, yprec = 8;
		if (xCol->isDouble())
			xprec *= 2;

		//
		// generate grace commands to setup labels
		//
		FileName fn(set->name());
		*os<<"@ title "<<'"'
		   <<(plotTitle.empty() ? fn.basename() : plotTitle)<<'"'<<endl;
		*os<<"@ subtitle "<<'"'<<table->name()<<'"'<<endl;
		*os<<"@ xaxis label "<<'"'<<xCol->name();
		if (offsetX)
			*os<<" - "<<setprecision(xprec)<<xOffset;

		if (!xCol->units().empty())
			*os<<" ["<<xCol->units()<<"]";
		*os<<'"'<<endl;

		//
		// if there are multiple y columns, produce a legend
		//
		if (plotVec.size() > 2)
			*os<<"@ legend on"<<endl;

		//
		//	scatter plot
		//
		vector<int>	setColors,
					setSymbols;
		bool scatter = booleanParameter("withscatter");
		if (scatter) {
			setColors = intListParameter("setcolors");
			setSymbols = intListParameter("setsymbols");
		}

		*os<<"@ with g0"<<endl;
		//
		//	axes properties
		//
		//AI (SOC-SPR-2536)
		//Also new versions of xmgrace need this keyword to
		//plot properly the axis, mainly if the user wants to 
		//change the min and max axis values.
		*os<<"@world "<<xWorldMin<<", "<<yWorldMin<<", "
			<<xWorldMax<<", "<< yWorldMax<<"\n";
		if (booleanParameter("withxmin"))
			*os<<"@  world xmin "<<realParameter("xmin")<<endl;
		if (booleanParameter("withxmax"))
			*os<<"@  world xmax "<<realParameter("xmax")<<endl;
		if (booleanParameter("withymin"))
			*os<<"@  world ymin "<<realParameter("ymin")<<endl;
		if (booleanParameter("withymax"))
			*os<<"@  world ymax "<<realParameter("ymax")<<endl;
		if (booleanParameter("xlog"))
			*os<<"@  xaxes scale Logarithmic\n";
		if (booleanParameter("ylog"))
			*os<<"@  yaxes scale Logarithmic\n";
		//AI (SOC-SPR-2536)
		//Needed in new xmgrace versions.
		*os<<"@    view 0.15, 0.15000, 0.85, 0.85\n";
		*os<<"@  autoticks\n";

		//
		// loop over all y columns
		//
		unsigned &i = plotIndex;
		for (; i<plotVec.size()-1; ++i) {
			const FixedSizeDataVector *yColData = plotVec.data(i+1);

			//
			//	set symbols/color in case of scatter plot
			//
			if (scatter) {
				int symbol = 8,	//	'+' in Grace
					color = 1;	//	black
				if (setSymbols.size() > i)
					symbol = setSymbols[i];
				if (setColors.size() > i)
					color = setColors[i];
				*os<<"@  s"<<i<<" symbol "<<symbol<<endl;
				*os<<"@  s"<<i<<" symbol size 1.0\n";
				*os<<"@  s"<<i<<" symbol color "<<color<<endl;
				*os<<"@  s"<<i<<" symbol pattern 1\n";
				*os<<"@  s"<<i<<" symbol fill pattern 0\n";
				*os<<"@  s"<<i<<" line type 0\n";
				//
				//	coordinate grid lines
				//
				*os<<"@  g0 xaxis tick major grid on\n";
				*os<<"@  g0 xaxis tick minor grid on\n";
				*os<<"@  g0 xaxis tick major linestyle 3\n";
				*os<<"@  g0 xaxis tick minor linestyle 3\n";
				*os<<"@  g0 yaxis tick major grid on\n";
				*os<<"@  g0 yaxis tick minor grid on\n";
				*os<<"@  g0 yaxis tick major linestyle 3\n";
				*os<<"@  g0 yaxis tick minor linestyle 3\n";
			}

			//
			// Produce the label for this column; if there is only one we can
			// use it for the y axis, otherwise we put it in the legend.
			//
			if (plotVec.size() == 2)
				*os<<"@ yaxis label ";
			else
				*os<<"@ s"<<i<<" legend ";
			*os<<'"'<<yColData->name();
			if (!yColData->units().empty())
				*os<<" ["<<yColData->units()<<"]";
			*os<<'"'<<endl;

			//	write 'type' keyword
			*os<<"@ type xy";
			bool hasXerr = plotVec.hasXerrors(i+1),
				 hasYerr = plotVec.hasYerrors(i+1);
			if (hasXerr)
				*os<<"dx";
			if (hasYerr)
				*os<<"dy";
			*os<<endl;

			//
			// print all x,y,[dx][dy] pairs of this column
			//
			if (yColData->isDouble())
				yprec *= 2;
			for (unsigned j=0; j<xCol->size(); ++j) {
				double xx = xCol->value(j),
					   yy = yColData->value(j);
		                if (!isnan(yy) && !isnan(xx)){
				 double x = xx, y = yy; 
				 if (NaN.isSet(x))
					continue;
				 *os<<setprecision(xprec)<<x-xOffset<<' '
				   <<setprecision(yprec)
				   <<(!NaN.isSet(y) ? y : 0.);
				 if (hasXerr) {
					double xerr = plotVec.xError(i+1)->value(j);
					*os<<' '<<(!NaN.isSet(xerr) ? xerr : 0.);
				 }
				 if (hasYerr) {
					double yerr = plotVec.yError(i+1)->value(j);
					*os<<' '<<(!NaN.isSet(yerr) ? yerr : 0.);
				 }
				 *os<<endl;
				} 
			}

			//
			// indicate the end of this set to grace
			//
			*os<<"&"<<endl;
		}
	}	//	after this block the DalFixedSizeDataVectors go out of scope
		//	in the destructors of which there is a delete on ColumnData *
		//	this *must not happen after* close(set)
	dataSetServer->close(set);
	}

	//
	//	process abscissa ranges
	//
	if (booleanParameter("withxranges")) {
		vector<string> tabList = stringListParameter("xrangetables");
		if (booleanParameter("withalltablesafter")) {
			//	all tables in set after table named in first element
			//	of 'xrangetables' are to be looked at
			if (tabList.empty())
				errstr<<"No initial abscissa range table specified"
					  <<Sas::error("NoAbscissaRangeTable");
			SetTableColumn stc(tabList[0]);
			tabList.clear();
			DataSet *set = stc.dataSet();
			unsigned fromBlock = set->blockNumber(stc.table()->name()),
					 toBlock = set->blocks();
			for (unsigned n=fromBlock; n<toBlock; ++n)
				tabList.push_back(set->block(n)->qualifiedName());

			dataSetServer->close(set);
		}

		AbscissaRanges ranges(tabList,
							  stringParameter("xrangefromcol"),
							  stringParameter("xrangetocol"));

		double deltaY = yWorldMax - yWorldMin,
			   // if there is no other data the d/f y-baseline for the
			   // ranges is 0 and the stride is .1				
			   yBase = deltaY > 0. ? yWorldMax + .1*deltaY : 0.,
			   yStride = deltaY > 0. ? .1*deltaY : .1;

		if (booleanParameter("withybase"))
			yBase = realParameter("ybase");
		if (booleanParameter("withystride"))
			yStride = realParameter("ystride");

		ranges.plot(yBase, yStride, offsetX, xOffset, plotIndex,
					deltaY <= 0., plotTitle, plotSubTitle, *os);
	}
}
